# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import csv
import collections
import sys
import xlwt

from .base import get_content

PY3 = sys.version_info[0] == 3


def convert(response, encoding='utf-8', title_sheet='Sheet 1',  content_attr='content', csv_kwargs=None):
    """Replace HttpResponse csv content with excel formatted data using xlwt
    library.
    """
    csv_kwargs = csv_kwargs or {}
    # Styles used in the spreadsheet.  Headings are bold.
    header_font = xlwt.Font()
    header_font.bold = True

    header_style = xlwt.XFStyle()
    header_style.font = header_font

    wb = xlwt.Workbook(encoding=encoding)
    ws = wb.add_sheet(title_sheet)

    # Cell width information kept for every column, indexed by column number.
    cell_widths = collections.defaultdict(lambda: 0)
    content = get_content(response)
    reader = csv.reader(content, **csv_kwargs)
    for lno, line in enumerate(reader):
        if lno == 0:
            style = header_style
        else:
            style = None
        write_row(ws, lno, line, cell_widths, style=style, encoding=encoding)
    # Roughly autosize output column widths based on maximum column size.
    for col, width in cell_widths.items():
        ws.col(col).width = width
    setattr(response, content_attr, '')
    wb.save(response)


def write_row(ws, lno, cell_text, cell_widths, style=None, encoding='utf-8'):
    """Write row of utf-8 encoded data to worksheet, keeping track of maximum
    column width for each cell.
    """
    import xlwt
    if style is None:
        style = xlwt.Style.default_style
    for cno, utf8_text in enumerate(cell_text):
        cell_text = utf8_text
        if not PY3:
            cell_text = cell_text.decode(encoding)
        ws.write(lno, cno, cell_text, style)
        cell_widths[cno] = max(cell_widths[cno],
                               get_xls_col_width(cell_text, style))


# A reasonable approximation for column width is based off zero character in
# default font.  Without knowing exact font details it's impossible to
# determine exact auto width.
# http://stackoverflow.com/questions/3154270/python-xlwt-adjusting-column-widths?lq=1
def get_xls_col_width(text, style):
    return int((1 + len(text)) * 256)



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import csv
import codecs
import sys

PY3 = sys.version_info[0] == 3

if PY3:
    string = str
    unicode = str
else:
    string = basestring

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import django_tables2 as tables

from django.conf import settings
from django.utils.translation import ugettext as _
from django.http import HttpResponse
from django.utils.html import strip_tags

from django_tables2_reports import csv_to_xls
from django_tables2_reports.utils import (DEFAULT_PARAM_PREFIX,
                                          get_excel_support,
                                          generate_prefixto_report)


# Unicode CSV writer, copied direct from Python docs:
# http://docs.python.org/2/library/csv.html

class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()
        self.encoding = encoding

    def writerow(self, row):
        if PY3:
            self.writer.writerow([s for s in row])
            # Fetch UTF-8 output from the queue ...
            data = self.queue.getvalue()
        else:
            self.writer.writerow([s.encode(self.encoding) for s in row])
            # Fetch UTF-8 output from the queue ...
            data = self.queue.getvalue()
            data = data.decode(self.encoding)
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)


class TableReport(tables.Table):

    def __init__(self, *args, **kwargs):
        if not 'template' in kwargs:
            kwargs['template'] = 'django_tables2_reports/table.html'
        prefix_param_report = kwargs.pop('prefix_param_report', DEFAULT_PARAM_PREFIX)
        super(TableReport, self).__init__(*args, **kwargs)
        self.param_report = generate_prefixto_report(self, prefix_param_report)
        self.formats = [(_('CSV Report'), 'csv')]
        if get_excel_support():
            self.formats.append((_('XLS Report'), 'xls'))

    def as_report(self, request, report_format='csv'):
        if report_format == 'csv':
            return self.as_csv(request)
        elif report_format == 'xls':
            return self.as_xls(request)
        raise ValueError("This format %s is not accepted" % report_format)

    def as_csv(self, request):
        response = HttpResponse()
        csv_writer = UnicodeWriter(response, encoding=settings.DEFAULT_CHARSET)
        csv_header = [column.header for column in self.columns]
        csv_writer.writerow(csv_header)
        for row in self.rows:
            csv_row = []
            for column, cell in row.items():
                if isinstance(cell, string):
                    # if cell is not a string strip_tags(cell) get an
                    # error in django 1.6
                    cell = strip_tags(cell)
                else:
                    cell = unicode(cell)
                csv_row.append(cell)
            csv_writer.writerow(csv_row)
        return response

    def as_xls(self, request):
        return self.as_csv(request)

    def treatement_to_response(self, response, report_format='csv'):
        if report_format == 'xls':
            csv_to_xls.convert(response, get_excel_support(),
                               encoding=settings.DEFAULT_CHARSET,
                               title_sheet=self.param_report[:csv_to_xls.MAX_LENGTH_TITLE_SHEET])
        return response



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from django_tables2_reports.utils import create_report_http_response, REQUEST_VARIABLE, REPORT_CONTENT_TYPE


class TableReportMiddleware(object):

    def process_response(self, request, response):
        table_to_report = getattr(request, REQUEST_VARIABLE, None)
        current_content_type = response.get('Content-Type', None)
        if table_to_report and current_content_type != REPORT_CONTENT_TYPE:
            return create_report_http_response(table_to_report, request)
        return response



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from django.http import HttpResponse
from django_tables2_reports import csv_to_xls

DEFAULT_PARAM_PREFIX = 'report'
REQUEST_VARIABLE = 'table_to_report'
REPORT_CONTENT_TYPE = 'application/vnd.ms-excel'
REPORT_MYMETYPE = REPORT_CONTENT_TYPE  # backwards compatible


def create_report_http_response(table, request):
    report_format = request.GET.get(table.param_report)
    report = table.as_report(request, report_format=report_format)
    extension = get_extension_report(report_format)
    filename = '%s.%s' % (table.param_report, extension)
    response = HttpResponse(report, content_type=REPORT_CONTENT_TYPE)
    response['Content-Disposition'] = 'attachment; filename=%s' % filename
    response = table.treatement_to_response(response, report_format=report_format)
    return response


def get_excel_support():
    # If you don't specify a xls library, this function will autodetect the library to use for xls writing. Default to xlwt.
    from django.conf import settings
    return getattr(settings, "EXCEL_SUPPORT", None) or csv_to_xls.get_xls_support()


def get_extension_report(report_format):
    if report_format == 'xls' and get_excel_support() == "openpyxl":
        return 'xlsx'
    return report_format


def generate_prefixto_report(table, prefix_param_report=None):
    param_report = prefix_param_report or DEFAULT_PARAM_PREFIX
    table_class = table.__class__
    prefix = table.prefix
    param_report = "%s-%s" % (param_report, table_class.__name__.lower())
    if prefix:
        param_report = "%s-%s" % (prefix, param_report)
    return param_report



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from django_tables2.config import RequestConfig

from django_tables2_reports.utils import REQUEST_VARIABLE


class RequestConfigReport(RequestConfig):

    def __init__(self, request, paginate=True, paginate_report=False):
        self.request = request
        self.paginate = paginate
        self.paginate_report = paginate_report

    def configure(self, table, extra_context=None):
        table.is_configured = True
        param_report = table.param_report
        is_report = self.request.GET.get(param_report)
        table_to_report = None
        if is_report:
            self.paginate = self.paginate_report
            table_to_report = table
            setattr(self.request, REQUEST_VARIABLE, table_to_report)
            self.request.extra_context = extra_context or {}
        super(RequestConfigReport, self).configure(table)
        return table_to_report



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import sys

from django.conf import settings
from django.http import HttpRequest
from django.test import TestCase
from django.utils.unittest import skipIf

import django_tables2
import django_tables2_reports.tables
import django_tables2_reports.views


PY3 = sys.version_info[0] == 3
if PY3:
    unichr = chr


class TableReportForTesting(django_tables2_reports.tables.TableReport):
    name = django_tables2.Column()
    item_num = django_tables2.Column()


class ReportTableViewForTesting(django_tables2_reports.views.ReportTableView):
    table_class = TableReportForTesting


class TestCsvGeneration(TestCase):
    """Test csv generation on sample table data."""

    def test_csv_simple_input(self):
        """Test ability to generate csv with simple input data."""

        # Mix of integer and string data.  Ensure that commas and
        # quotes are escaped properly.
        data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': 'String, with, commas',
                'item_num': 2,
            },
            {
                'name': 'String with " quote',
                'item_num': 3,
            },
        ]

        table = TableReportForTesting(data)
        response = table.as_csv(HttpRequest())
        self.assertEqual(response.status_code, 200)
        # Expect cells containing commas to be escaped with quotes.
        content = response.content
        if PY3:
            content = content.decode(settings.DEFAULT_CHARSET).replace('\x00', '')
        self.assertEqual(
            content,
            'Name,Item Num\r\n'
            'Normal string,1\r\n'
            '"String, with, commas",2\r\n'
            '"String with "" quote",3\r\n')

    def test_csv_with_unicode(self):
        """Test that unicode cell values are converted correctly to csv."""

        data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': u'String with ' + unichr(0x16c) + ' char',
                'item_num': 2,
            },
        ]

        table = TableReportForTesting(data)
        response = table.as_csv(HttpRequest())
        self.assertEqual(response.status_code, 200)
        # Expect csv content to be utf-8 encoded.
        content = response.content
        result = ('Name,Item Num\r\n'
                  'Normal string,1\r\n'
                  'String with ' + unichr(0x16c) + ' char,2\r\n')
        if PY3:
            content = content.decode(settings.DEFAULT_CHARSET).replace('\x00', '')
        else:
            result = result.encode(settings.DEFAULT_CHARSET)
        self.assertEqual(content, result)

    def test_csv_no_pagination(self):
        """Ensure that table pagination doesn't affect output."""

        data = [
            {
                'name': 'page 1',
                'item_num': 1,
            },
            {
                'name': 'page 2',
                'item_num': 2,
            },
        ]

        table = TableReportForTesting(data)
        table.paginate(per_page=1)

        response = table.as_csv(HttpRequest())
        self.assertEqual(response.status_code, 200)
        # Ensure that even if table paginated, output is all row
        # data.
        content = response.content
        if PY3:
            content = content.decode(settings.DEFAULT_CHARSET).replace('\x00', '')
        self.assertEqual(
            content,
            ('Name,Item Num\r\n'
             'page 1,1\r\n'
             'page 2,2\r\n')
        )


@skipIf(
    not django_tables2_reports.utils.get_excel_support(),
    "No Excel support, please install xlwt, pyExcelerator or openpyxl")
class TestExcelGeneration(TestCase):
    def setUp(self):
        # Mix of integer and string data.  Ensure that commas and
        # quotes are escaped properly.
        self.data = [
            {
                'name': 'Normal string',
                'item_num': 1,
            },
            {
                'name': 'String, with, commas',
                'item_num': 2,
            },
            {
                'name': 'String with " quote',
                'item_num': 3,
            },
            {
                'name': u'String with ' + unichr(0x16c) + ' char',
                'item_num': 4,
            }
        ]
        self.table = TableReportForTesting(self.data)

    def test_excel_simple_input(self, extension='xls'):
        """Test ability to generate excel output with simple input data."""
        excel_support = getattr(settings, 'EXCEL_SUPPORT', django_tables2_reports.utils.get_excel_support())
        response = self.table.treatement_to_response(
            self.table.as_csv(HttpRequest()),
            report_format='xls')
        self.assertEqual(response.status_code, 200)
        open('test-file-%s.%s' % (excel_support, extension),
             'wb').write(response.content)

    def test_pyexcelerator(self):
        if PY3:
            return
        settings.EXCEL_SUPPORT = "pyexcelerator"
        self.test_excel_simple_input()

    def test_xlwt(self):
        settings.EXCEL_SUPPORT = "xlwt"
        self.test_excel_simple_input()

    def test_openpyxls(self):
        settings.EXCEL_SUPPORT = "openpyxl"
        self.test_excel_simple_input(extension='xlsx')



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.


from django_tables2.views import SingleTableView

from django_tables2_reports.config import RequestConfigReport
from django_tables2_reports.utils import create_report_http_response


class ReportTableView(SingleTableView):

    def get_table(self):
        """
        Return a table object to use. The table has automatic support for
        sorting and pagination.
        """
        options = {}
        table_class = self.get_table_class()
        table = table_class(self.get_table_data())
        paginate = self.get_table_pagination()
        if paginate is not None:
            options['paginate'] = paginate
        self.table_to_report = RequestConfigReport(self.request, **options).configure(table)
        return table

    def render_to_response(self, context, **response_kwargs):
        if self.table_to_report:
            return create_report_http_response(self.table_to_report, self.request)
        return super(ReportTableView, self).render_to_response(context, **response_kwargs)



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import sys

PY3 = sys.version_info[0] == 3

if PY3:
    from io import StringIO
else:
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO


def get_content(file_or_response, encoding='utf-8', content_attr='content'):
    content = getattr(file_or_response, content_attr)
    if PY3 and isinstance(content, bytes):
        content = content.decode(encoding).replace('\x00', '')
    return StringIO(content)



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

#http://stackoverflow.com/questions/3681868/is-there-a-limit-on-an-excel-worksheets-name-length
MAX_LENGTH_TITLE_SHEET = 31


def convert(response, excel_support=None, encoding='utf-8',
            title_sheet='Sheet 1', content_attr='content', csv_kwargs=None):
    if len(title_sheet) > MAX_LENGTH_TITLE_SHEET:
        raise ValueError("The maximum length of a title of a sheet is %s" % MAX_LENGTH_TITLE_SHEET)
    excel_support = excel_support or get_xls_support()
    if excel_support == 'xlwt':
        from .xlwt_converter import convert
    elif excel_support == 'openpyxl':
        from .openpyxl_converter import convert
    elif excel_support == 'pyexcelerator':
        from .pyexcelerator_converter import convert
    else:
        raise RuntimeError("No support for xls generation available")

    convert(response,
            encoding=encoding,
            title_sheet=title_sheet,
            content_attr=content_attr,
            csv_kwargs=csv_kwargs)


def get_xls_support():
    try:
        import xlwt
        return 'xlwt'
    except ImportError:
        pass
    try:
        import openpyxl
        return 'openpyxl'
    except ImportError:
        pass
    try:
        import pyExcelerator
        return 'pyexcelerator'
    except ImportError:
        pass
    return None



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import csv
import collections
import sys

from openpyxl import Workbook
from openpyxl.cell import get_column_letter

from .base import get_content

PY3 = sys.version_info[0] == 3


def convert(response, encoding='utf-8', title_sheet='Sheet 1', content_attr='content', csv_kwargs=None):
    csv_kwargs = csv_kwargs or {}
    wb = Workbook(encoding=encoding)
    ws = wb.get_active_sheet()
    ws.title = title_sheet
    cell_widths = collections.defaultdict(lambda: 0)
    content = get_content(response, encoding=encoding, content_attr=content_attr)
    reader = csv.reader(content, **csv_kwargs)

    for lno, line in enumerate(reader):
        write_row(ws, lno, line, cell_widths, encoding=encoding)

    # Roughly autosize output column widths based on maximum column size
    # and add bold style for the header
    for i, cell_width in cell_widths.items():
        ws.cell(column=i, row=0).style.font.bold = True
        ws.column_dimensions[get_column_letter(i + 1)].width = cell_width

    setattr(response, content_attr, '')
    wb.save(response)


def write_row(ws, lno, cell_text, cell_widths, encoding='utf-8'):
    for cno, cell_text in enumerate(cell_text):
        if not PY3:
            cell_text = cell_text.decode(encoding)
        ws.cell(column=cno, row=lno).value = cell_text
        cell_widths[cno] = max(
            cell_widths[cno],
            len(cell_text))



# -*- coding: utf-8 -*-
# Copyright (c) 2012-2013 by Pablo Martín <goinnn@gmail.com>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

# This only works with python 2.x

import csv
import pyExcelerator


from .base import get_content


def convert(response, encoding='utf-8', title_sheet='Sheet 1', content_attr='content', csv_kwargs=None):
    csv_kwargs = csv_kwargs or {}
    workbook = pyExcelerator.Workbook()
    worksheet = workbook.add_sheet(title_sheet)
    lno = 0
    content = get_content(response)
    reader = csv.reader(content, **csv_kwargs)
    for line in reader:
        write_row(worksheet, lno, line, encoding=encoding)
        lno = lno + 1
    setattr(response, 'content_attr', workbook.get_biff_data())


def write_row(worksheet, lno, columns, encoding='utf-8'):
    """ Write a non-header row into the worksheet """
    cno = 0
    for column in columns:
        worksheet.write(lno, cno, column.decode(encoding))
        cno = cno + 1
